package minischeme

import TreeGenerator._

object EtaExpander extends (List[Tree] => List[Tree]) {
  def apply(trees: List[Tree]): List[Tree] =
    trees map transform

  private def transform(t: Tree): Tree = copyPos(t) {
    t match {
      case SList(SIdent("define") :: name :: body) =>
        SList(SIdent("define") :: name :: (body map transform))
      case SList(SIdent("lambda") :: args :: body) =>
        SList(SIdent("lambda") :: args :: (body map transform))
      case SList(SIdent("let") :: SList(bindings) :: body) =>
        val newBindings = bindings map {
          case SList(name :: value :: Nil) =>
            SList(name :: transform(value) :: Nil)
        }
        SList(SIdent("let") :: SList(newBindings) :: (body map transform))
      case SList((fun @ SIdent(name)) :: args) if Primitives.is(fun) =>
        SList(fun :: (args map transform))
      case SList(elems) =>
        SList(elems map transform)
      case (ident @ SIdent(name)) if Primitives.is(ident) =>
        etaExpand(ident)
      case SList(Nil) | SIdent(_) | SNum(_) =>
        t
    }
  }

  private def etaExpand(prim: SIdent): Tree = {
    val arity = Primitives.arity(prim.name)
    if (arity < 0)
      Report.error(prim.pos,
                   "primitive " + prim.name + " cannot be eta-expanded!")
    val pSyms = List.tabulate(arity, {n: Int => new Symbol("x" + n, LocalKind)})
    lambda(pSyms map ident, List(SList(prim :: (pSyms map ident))))
  }

  private def copyPos(srcTree: Tree)(dstTree: Tree) = {
    dstTree.pos = srcTree.pos
    dstTree
  }
}
