package minischeme

import java.io._
import TreeGenerator._

/**
 * Parser for minischeme.
 *
 * @author Michel Schinz
 */
class Parser(scanner: Scanner) {
  var token: Token = scanner.nextToken
  def skipToken(): Unit =
    token = scanner.nextToken

  // program ::= { expr }
  def program(): List[Tree] = listOfExpr()

  // expr ::= list | symbol | number
  private def expr(): Tree = setPos { token match {
    case LeftParen => list()
    case IdentToken(n) => skipToken(); SIdent(n)
    case NumberToken(v) => skipToken(); SNum(v)
    case StringToken(s) =>
      skipToken()
      vector(SNum(s.length) :: (s.toList map { x: Char => SNum(x) }))
    case _ =>
      Report.error(token.pos, "unexpected token " + token)
  }}

  // list ::= '(' { expr } ')'
  private def list(): Tree = setPos {
    skipToken()
    val res = desugar(listOfExpr)
    if (token != RightParen)
      Report.error(token.pos, "expected ')', found " + token)
    skipToken()
    SList(res)
  }

  private def listOfExpr(): List[Tree] = token match {
    case LeftParen | NumberToken(_) | IdentToken(_) | StringToken(_) =>
      val head = expr()
      val tail = listOfExpr()
      head :: tail
    case _ =>
      List()
  }

  private def setPos(body: =>Tree): Tree = {
    val tok = token
    val tree = body
    tree.pos = tok.pos
    tree
  }

  private def desugar(lst: List[Tree]): List[Tree] = lst match {
    case SIdent("and") :: c1 :: c2 :: Nil =>
      SIdent("if") :: c1 :: c2 :: SNum(0) :: Nil
    case SIdent("or") :: c1 :: c2 :: Nil =>
      SIdent("if") :: c1 :: SNum(1) :: c2 :: Nil
    case _ => lst
  }
}
